{//////////////////////////////////////////////////////////////////////////}
{//                                                                      //}
{//  TYPE DEFINITIONS for GPC; BP compatibility and other things         //}
{//                                                                      //}
{//  v1.00 - by Dr Abimbola A. Olowofoyeku (The African Chief);          //}
{//                                                                      //}
{//  This package is released as part of the GNU Pascal project.         //}
{//                                                                      //}
{//  This library is a free library; you can redistribute and/or modify  //}
{//  it under the terms of the GNU Library General Public License,       //}
{//  version 2, as published by the Free Software Foundation.            //}
{//                                                                      //}
{//  This library is distributed in the hope that it will be useful,     //}
{//  but WITHOUT ANY WARRANTY; without even the implied warranty of      //}
{//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       //}
{//  GNU Library General Public License for more details.                //}
{//                                                                      //}
{//  You should have received a copy of the GNU Library General Public   //}
{//  License along with this library (see the file COPYING.LIB);         //}
{//  if not, write to the Free Software Foundation, Inc., 675 Mass Ave,  //}
{//  Cambridge, MA 02139, USA.                                           //}
{//                                                                      //}
{//  v1.00, May 1997 - Dr A Olowofoyeku: laa12@keele.ac.uk               //}
{//  Homepage: http://ourworld.compuserve.com/homepages/African_Chief    //}
{//////////////////////////////////////////////////////////////////////////}


{$I SYSTEM.INC}   { include the SYSTEM include file }

UNIT GPCTYPES;

INTERFACE

{
    Unsigned = positive values only
    Signed   = both positive and negative values

    Byte		unsigned	 8 bits
    ByteInt		signed		 8 bits

    ShortWord		unsigned	16 bits
    Word		unsigned	32 bits
    LongWord		unsigned	64 bits

    ShortInt		signed		16 bits
    Integer		signed		32 bits
    LongInt		signed		64 bits

    Single              = ShortReal	32 Bits
    Double              = Real		64 Bits
    Extended            = LongReal	80 Bits on the PC,
                                        maybe 128 elsewhere
}


Const
MaxStrSize   = 259;  {max size for normal strings}
FD_SETSIZE   = 256;
NULL         = 0;
bpStringSize = 255; {max size for Borland strings}

{ file constants - needed by SYSTEM.PAS and others }
Const L_READONLY   = $0000;
Const L_WRITEONLY  = $0001;
Const L_READWRITE  = $0002;
Const L_ACCESSMODE = $0003;
Const L_IWRITE 	   = 00200;
Const L_IREAD      = 00400;

Const LSH_COMPAT   = $0000;
Const LSH_DENYRW   = $0010;
Const LSH_DENYWR   = $0020;
Const LSH_DENYRD   = $0030;
Const LSH_DENYNO   = $0040;

Const L_BINARY     = $0004; {/* must fit in char, reserved by dos */}
Const L_TEXT       = $0008; {/* must fit in char, reserved by dos */}
Const L_NOINHERIT  = $0080; {/* DOS-specific */}

{ file naming constants }
Const
{$ifdef _Borland_16_Bit_}
fsPathName   = 80;
fsDirectory  = 66;
fsFileName   = 9;
fsExtension  = 5;
{$else _Borland_16_Bit_}
fsPathName   = MaxStrSize;
fsDirectory  = MaxStrSize;
fsFileName   = MaxStrSize;
fsExtension  = MaxStrSize;
{$endif _Borland_16_Bit_}

{ string types }
Type
HugeString = String ( MaxInt - 32 );
BigString  = String ( 1024 * 64 );  { 64kb capacity }
MidString  = String ( 1024 * 32 );  { 32kb capacity }
SmallString= String ( 1024 * 16 );  { 16kb capacity }
String1024 = String ( 1024 );       { 1kb  capacity }
String256  = String ( 256 );
NormString = String [ MaxStrSize ];
bpString   = String [ bpStringSize ];
PathStr	   = String [ fsPathName ];
DirStr     = String [ fsDirectory ];
NameStr    = String [ fsFileName ];
ExtStr     = String [ fsExtension ];
{$ifdef _Borland_Pascal_}
String     = bpString;
{$endif _Borland_Pascal_}

{ 0-based array of Chars }
Type
TCharArray = Packed Array [ 0 .. MaxInt-31 ] of Char;
pCharArray = ^TCharArray;

{ array of bytes }
TByteArray = Array [ 1 .. MaxInt-32 ] of Byte;
pByteArray = ^TByteArray;

{ array of words }
TWordArray = Array [ 1 .. MaxInt-32 ] of Word;
pWordArray = ^TWordArray;

{ array of integers }
TIntArray = Array [ 1 .. MaxInt-32 ] of Integer;
pIntArray = ^TIntArray;

{ array of longints }
TLongArray = Array [ 1 .. MaxInt-32 ] of Longint;
pLongArray = ^TLongArray;

{ array of reals }
TRealArray = Array [ 1 .. MaxInt-32 ] of Real;
pRealArray = ^TRealArray;

{ array of doubles }
TDoubleArray = Array [ 1 .. MaxInt-32 ] of Double;
pDoubleArray = ^TDoubleArray;

{ other types }
Type
size_t   = LongWord;
time_t   = Word;
clock_t  = integer;
gid_t    = integer;
off_t    = integer;
pid_t    = integer;
ssize_t  = integer;
dev_t    = integer;
ino_t    = integer;
mode_t   = integer;
nlink_t  = integer;
uid_t    = integer;
wchar_t  = integer;
va_list  = ^Void;

Type
ByteInt   = __byte__ Integer;
Byte      = __unsigned__ ByteInt;
ByteFile  = File of Byte;
VarArgs   = pChar;
Thandle   = integer;

{ Pointer types }
Type
Pointer   = ^Void;
pVoid     = Pointer;
pPointer  = ^Pointer;
pChar     = cString;
ppChar    = ^pChar;
pInteger  = ^integer;
pWord     = ^Word;
pLongint  = ^Longint;
pByte     = ^Byte;
pReal     = ^Real;
pDouble   = ^Double;
pString   = ^String;

{ procedural types }
Type
pFunction  = ^Function : Integer;
FuncPtr    = Function  : Integer;

pProcedure = ^Procedure;
ProcPtr    = Procedure;

{$ifdef _Borland_16_Bit_}
Const
pi = 3.14159265358979323846;
MaxLongInt = MaxInt;

Type
Single    = __short__ Real;
Double    = Real;          {* !!! There is no 6-Byte Real *}
Extended  = __long__ Real;
ShortInt  = __byte__ Integer;
LongInt   = Integer;
Comp      = __longlong__ Integer;
Integer16 = __short__ Integer;
Word      = __unsigned__ Integer16;
{$endif _Borland_16_Bit_}

IMPLEMENTATION

end.
